from inspect import Signature, signature

from fastapi import FastAPI
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import get_dependant

from fastapi_better_di._utils import current_app
from fastapi_better_di.patcher._utils import copy_func, decorate_method


def is_pathed() -> bool:
    return all(
        (
            getattr(FastAPI.__init__, "_IS_PATHED", False),
            getattr(get_dependant, "_IS_PATHED", False),
        )
    )


def patch():
    FastAPI.__init__ = decorate_method(
        FastAPI.__init__, after_call=lambda self, *_, **__: current_app.set(self)
    )

    get_dependant.__original__ = copy_func(get_dependant)
    get_dependant.__code__ = get_dependant_patched.__code__
    get_dependant.__signature__ = signature(get_dependant.__original__)
    get_dependant.__cache__ = {}

    get_dependant._IS_PATHED = True


def get_dependant_patched(*args, **kwargs) -> Dependant:
    cache_key = hash(kwargs["call"])

    if cache_key in get_dependant.__cache__:  # type: ignore
        return get_dependant.__cache__[cache_key]  # type: ignore

    from fastapi_better_di.exeptions import EarlyInit
    from fastapi_better_di.patcher._utils import patch_endpoint_handler
    from fastapi_better_di._utils import current_app

    get_dependant_original = get_dependant.__original__  # type: ignore

    sig: Signature = get_dependant_original.__signature__

    default_args = {
        k: v.default for k, v in sig.parameters.items() if v.default is not sig.empty
    }

    kwargs = {**default_args, **kwargs}

    try:
        app = current_app.get()
    except LookupError:
        raise EarlyInit("The main app must be initialized before importing routers")

    patch_endpoint_handler(kwargs["call"], app.dependency_overrides)
    result = get_dependant_original(*args, **kwargs)

    get_dependant.__cache__[cache_key] = result  # type: ignore

    return result



import os

from setuptools import find_packages, setup

VERSION = "0.3.0"


def get_readme() -> str:
    """Load the contents of the README file"""
    readme_path = os.path.join(os.path.dirname(__file__), "README.md")
    with open(readme_path, "r") as f:
        return f.read()


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="fastapi-better-di",
    description="",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    version=VERSION,
    packages=find_packages(include=["fastapi_better_di", "fastapi_better_di.*"]),
    url="https://github.com/MaximZayats/fastapi-better-di",
    author="Maxim",
    author_email="maximzayats1@gmail.com",
    install_requires=get_requirements(),
    keywords=[
        "python",
        "fastapi",
        "di",
        "dependency injection",
        "fastapi di",
    ],
)



from inspect import signature
from types import FunctionType
from typing import Callable, Dict, Optional

from fastapi.params import Depends


def patch_endpoint_handler(
    func: Callable, dependencies: Dict[Callable, Callable]
) -> None:
    if getattr(func, "_IS_PATCHED", None) is not None:
        return

    sig = signature(func)

    for parameter in sig.parameters.values():
        if parameter.annotation in dependencies:
            if isinstance(parameter.default, Depends):
                if parameter.default.dependency is not None:
                    return
                # def func(param: Type = Depends())
                dependency = dependencies[parameter.annotation]
                parameter._default.dependency = dependency  # NOQA
            elif parameter.default == parameter.empty:
                # def func(param: Type)
                dependency = dependencies[parameter.annotation]
                parameter._default = Depends(dependency)  # NOQA

    func.__signature__ = sig
    func._IS_PATCHED = True


def decorate_method(
    method: Callable,
    before_call: Optional[Callable] = None,
    after_call: Optional[Callable] = None,
) -> Callable:
    def wrapper(*args, **kwargs):
        if before_call is not None:
            before_call(*args, **kwargs)

        result = method(*args, **kwargs)

        if after_call is not None:
            after_call(*args, **kwargs)

        return result

    wrapper._IS_PATHED = True

    return wrapper


def copy_func(f, name=None) -> FunctionType:
    fn = FunctionType(
        f.__code__, f.__globals__, name or f.__name__, f.__defaults__, f.__closure__
    )

    fn.__signature__ = signature(f)
    fn.__dict__.update(f.__dict__)

    return fn
